# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
# Author: Daniel Walton, dwalton@cumulusnetworks.com
#

import logging
import os
import re
import sre_constants
import sys
import subprocess
from collections import OrderedDict
from copy import deepcopy, copy
from ipaddr import IPv4Address, IPv6Address, IPv4Network, IPv6Network, IPAddress, IPNetwork
from pprint import pformat

log = logging.getLogger(__name__)

# Colors for printing
red    = 91
green  = 92
yellow = 93
blue   = 94

loose_matching_variables = ['<text>', '<interface>', '<wildcard>', '<regex>', '']
custom_matchers = {}


def is_loose_matching_variable(word):
    # There are variable types that will match on just about anything...<text>, <wildcard>, etc.
    # Return True if this is one of those variable types
    if word in loose_matching_variables:
         return True

    # <text-foo> counts as loose because <text> is loose
    for x in loose_matching_variables:
        # Chop the > at the end and add a '-'
        x = x[0:-1] + '-'
        if word.startswith(x):
            return True

    return False


def get_bash_prompt():
    """
    This takes 40-50ms...we need to find a better/faster way to get the prompt
    """
    try:
        subprocess.check_output("/bin/echo /bin/false | /bin/bash -i",
                                stderr=subprocess.STDOUT, shell=True)
        return None
    except subprocess.CalledProcessError as e:
        return e.output.splitlines()[0].rstrip().replace('/bin/false', '')


def get_network_docopt_info(argv):

    # Note if the last arg was '--completions'
    print_options = False
    ended_with_space = False
    len_argv = len(argv)

    if len_argv >= 2 and argv[-2] == '--completions':
        print_options = True
        init_completion_cword = int(argv[-1])

        # If the user did:
        #   "iface swp1 add <TAB>"
        #
        # then argv will be:
        #   ['./iface', 'swp1', 'add', '--completions', '3']
        #
        # but if they did
        #   "iface swp1 add<TAB>"
        #
        # then argv will be:
        #   ['./iface', 'swp1', 'add', '--completions', '2']
        #
        # We need to know how many keywords they entered so
        # subtract 1 for 'iface' and 2 for '--completions 2'
        len_argv_pre_tab_complete = len_argv - 3

        if init_completion_cword > len_argv_pre_tab_complete:
            ended_with_space = True
        else:
            ended_with_space = False

        argv = argv[0:-2]

    return (print_options, ended_with_space, argv)


def sort_for_humans(list_foo):
    """
    Sort list_foo in the way that humans expect
    http://nedbatchelder.com/blog/200712/human_sorting.html
    """
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    list_foo.sort(key=alphanum_key)
    return list_foo


def sort_ifaces(ifnames):
    """
    Given a list of interface names, return a list of those names sorted
    in the following order:
        lo -> eth -> swp -> bonds, bridges, etc
    """
    sorted_interfaces_lo = []
    sorted_interfaces_eth = []
    sorted_interfaces_swp = []
    sorted_interfaces_nonswp = []

    for x in sort_for_humans(ifnames):
        if x.startswith('lo'):
            sorted_interfaces_lo.append(x)

        elif x.startswith('eth'):
            sorted_interfaces_eth.append(x)

        elif x.startswith('swp'):
            sorted_interfaces_swp.append(x)

        else:
            sorted_interfaces_nonswp.append(x)

    # print "sorted_interfaces_lo %s" % sorted_interfaces_lo
    # print "sorted_interfaces_eth %s" % sorted_interfaces_eth
    # print "sorted_interfaces_swp %s" % sorted_interfaces_swp
    # print "sorted_interfaces_nonswp %s" % sorted_interfaces_nonswp

    return (sorted_interfaces_lo +
            sorted_interfaces_eth +
            sorted_interfaces_swp +
            sorted_interfaces_nonswp)


def soundex(name, length=12):
    """
    http://code.activestate.com/recipes/52213-soundex-algorithm/

    soundex module conforming to Knuth's algorithm
    implementation 2000-12-24 by Gregory Jorgensen
    public domain
    """
    # digits holds the soundex values for the alphabet
    digits = '01230120022455012623010202'
    sndx = ''
    fc = ''

    # translate alpha chars in name to soundex digits
    for c in name.upper():
        if c.isalpha():
            if not fc:
                fc = c   # remember first letter

            d = digits[ord(c) - ord('A')]

            # duplicate consecutive soundex digits are skipped
            if not sndx or (d != sndx[-1]):
                sndx += d

    # replace first digit with first alpha character
    sndx = fc + sndx[1:]

    # remove all 0s from the soundex code
    sndx = sndx.replace('0', '')

    # pad soundex code to 'length' characters
    sndx = (sndx + (length * '0'))[:length]

    return sndx


def swap_letter_pairs(name):
    """
    If name is 'zelda' return
        ['ezlda', 'zleda', 'zedla', 'zelad']
    """
    names = []
    name = list(name)

    for index in xrange(len(name) - 1):
        char_at_index = name[index]
        char_at_index_plus_one = name[index+1]
        tmp = copy(name)
        tmp[index] = char_at_index_plus_one
        tmp[index+1] = char_at_index
        names.append(''.join(tmp))
    return names


def drop_each_letter(name):
    """
    If name is 'zelda' return
        ['elda', 'zlda', 'zeda', 'zela']
    """
    names = []
    name = list(name)

    for index in xrange(len(name) - 1):
        tmp = copy(name)
        del tmp[index]
        names.append(''.join(tmp))
    return names


def interface_yes_no_list(argv_word, yes_no_auto_etc):
    """
    Used for matching swp=value lists like this one:
    swp1=yes swp2=no swp3=1 swp4=0
    """

    # These are all commands that can be configured via a list under a bridge
    # interface or as just a "keyword (yes|no)" under the interface
    if argv_word in yes_no_auto_etc:
        return True

    all_match = True

    for x in argv_word.replace(',', ' ').strip().split():
        x = x.strip()

        if '=' in x:
            (interface, value) = x.split('=')
            interface = interface.strip()
            value = value.strip()

            if value not in yes_no_auto_etc:
                all_match = False
                break
        else:
            all_match = False
            break

    return all_match


class Word(object):
    """
    A keyword:
    - bgp
    - show
    - neighbor

    or a variable:
    - <ipv4>
    - <ipv4/prefixlen>
    - <mac>
    """

    def __init__(self, text):
        self.text = text
        self.value = None
        self.value_raw = None
        self.fuzzy = False
        self.len_text = len(self.text)

        for x in ('[', ']', '(', ')', '|'):
            assert x not in self.text, "%s contains %s, nested expressions are not supported" % (self.text, x)

        # if self.text.startswith('<'):
        #     self.soundex = None
        # else:
        #     self.soundex = soundex(self.text)

    def __str__(self):
        return self.text

    def reset(self):
        self.value = None
        self.fuzzy = False

    def match(self, argv_word, use_fuzzy, tab_completing):
        """
        Return True if argv_word is a match
        """
        if self.text.startswith('<'):

            # Our client may want to use a new <foo> tag that has some special meaning to them.
            # We do not want to clutter network docopt with support for <foo> so instead we let
            # the client provide a hook to a function that can be used to tell us if
            if self.text in custom_matchers:
                matcher = custom_matchers[self.text]

                if matcher(argv_word, False, tab_completing):
                    self.value = argv_word
                    self.value_raw = argv_word
                    return True

                if use_fuzzy and matcher(argv_word, True, tab_completing):
                    self.value = argv_word
                    self.value_raw = argv_word
                    self.fuzzy = True
                    return True

            # "1.1.1.1 vrf FOO" but the "vrf FOO" is optional
            elif self.text == '<ipv4-vrf-text>':
                foo = argv_word.strip().split()

                if len(foo) == 1:
                    first_word = foo[0]

                    try:
                        IPv4Address(first_word)
                        self.value = argv_word
                        self.value_raw = argv_word
                        return True
                    except ValueError:
                        pass

                elif len(foo) >= 2:
                    first_word = foo[0]
                    second_word = foo[1]

                    if second_word == 'vrf' and '/' not in first_word:
                        try:
                            IPv4Address(first_word)
                            self.value = argv_word
                            self.value_raw = argv_word
                            return True
                        except ValueError:
                            pass

            # ipv4 or ipv6 address
            elif self.text == '<ip>' or self.text.startswith('<ip-'):
                try:
                    if '/' not in argv_word and IPAddress(argv_word):
                        self.value = argv_word
                        self.value_raw = argv_word
                        return True
                except ValueError:
                    pass

                if use_fuzzy:
                    try:
                        if '/' in argv_word and IPNetwork(argv_word):
                            self.value = argv_word
                            self.value_raw = argv_word
                            self.fuzzy = True
                            return True
                    except ValueError:
                        pass

            # ipv4 or ipv6 address with prefix length
            elif self.text == '<ip/prefixlen>':
                try:
                    if '/' in argv_word and IPNetwork(argv_word):
                        self.value = argv_word
                        self.value_raw = argv_word
                        return True
                except ValueError:
                    pass

                if use_fuzzy:
                    try:
                        if '/' not in argv_word and IPAddress(argv_word):
                            self.value = argv_word
                            self.value_raw = argv_word
                            self.fuzzy = True
                            return True
                    except ValueError:
                        pass

            elif self.text == '<ipv4>' or self.text.startswith('<ipv4-'):
                try:
                    if '/' not in argv_word and IPv4Address(argv_word):
                        self.value = argv_word
                        self.value_raw = argv_word
                        return True
                except ValueError:
                    pass

                if use_fuzzy:
                    try:
                        if '/' in argv_word and IPv4Network(argv_word):
                            self.value = argv_word
                            self.value_raw = argv_word
                            self.fuzzy = True
                            return True
                    except ValueError:
                        pass

            elif self.text == '<ipv4/prefixlen>':
                try:
                    if '/' in argv_word and IPv4Network(argv_word):
                        self.value = argv_word
                        self.value_raw = argv_word
                        return True
                except ValueError:
                    pass

                if use_fuzzy:
                    try:
                        if '/' not in argv_word and IPv4Address(argv_word):
                            self.value = argv_word
                            self.value_raw = argv_word
                            self.fuzzy = True
                            return True
                    except ValueError:
                        pass

            elif self.text == '<ipv6>' or self.text.startswith('<ipv6-'):
                try:
                    if '/' not in argv_word and IPv6Address(argv_word):
                        self.value = argv_word
                        self.value_raw = argv_word
                        return True
                except ValueError:
                    pass

                if use_fuzzy:
                    try:
                        if '/' in argv_word and IPv6Network(argv_word):
                            self.value = argv_word
                            self.value_raw = argv_word
                            self.fuzzy = True
                            return True
                    except ValueError:
                        pass

            elif self.text == '<ipv6/prefixlen>':
                try:
                    if '/' in argv_word and IPv6Network(argv_word):
                        self.value = argv_word
                        self.value_raw = argv_word
                        return True
                except ValueError:
                    pass

                if use_fuzzy:
                    try:
                        if '/' not in argv_word and IPv6Address(argv_word):
                            self.value = argv_word
                            self.value_raw = argv_word
                            self.fuzzy = True
                            return True
                    except ValueError:
                        pass

            elif self.text == '<number-range-list>':
                all_digits_or_ranges = True
                for x in argv_word.split(','):
                    if x.isdigit() or re.match('^\d+-\d+$', x):
                        pass
                    else:
                        all_digits_or_ranges = False

                if all_digits_or_ranges:
                    self.value = argv_word.replace(',', ' ').strip()
                    self.value_raw = argv_word.replace(',', ' ').strip()
                    return True

            elif self.text in ('<wildcard>', '<regex>') or self.text.startswith('<wildcard-'):
                self.value = argv_word
                self.value_raw = argv_word
                return True

            elif self.text == '<mac>':
                # The first regex will match
                # 3D:F2:C9:A6:B3:4F
                # 3D-F2-C9-A6-B3-4F
                #
                # The second regex will match
                # 1234.abcd.1234
                if (re.match('(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}', argv_word) or
                    re.match('[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}', argv_word)):
                    self.value = argv_word
                    self.value_raw = argv_word
                    return True

            # 100=10.0.100.1 101=10.101.1
            elif self.text == '<number-ipv4-list>':
                all_match = True

                for x in argv_word.replace(',', ' ').strip().split():
                    if '=' in x:
                        (interface, value) = x.split('=')
                        interface = interface.strip()
                        value = value.strip()

                        if '/' in value:
                            all_match = False
                            break
                        else:
                            try:
                                IPv4Address(x)
                            except ValueError:
                                all_match = False
                                break
                    else:
                        all_match = False
                        break

                if all_match:
                    self.value = argv_word.replace(',', ' ').strip()
                    self.value_raw = argv_word.replace(',', ' ').strip()
                    return True

            # swp1 swp2 swp3
            elif self.text == '<interface-list>':
                self.value = argv_word.replace(',', ' ').strip()
                self.value_raw = argv_word.replace(',', ' ').strip()
                return True

            # swp1=100 swp2=100
            elif self.text == '<interface-range-list>':
                all_match = True

                for x in argv_word.replace(',', ' ').strip().split():
                    if '=' in x:
                        (interface, value) = x.split('=')
                        interface = interface.strip()
                        value = value.strip()

                        if not value.isdigit():
                            all_match = False
                            break

                    else:
                        all_match = False
                        break

                if all_match:
                    self.value = argv_word.replace(',', ' ').strip()
                    self.value_raw = argv_word.replace(',', ' ').strip()
                    return True

            # 00:11:22:33:44:01 10.0.0.1/24 10.0.1.2/24
            elif self.text == '<mac-ip/prefixlen-list>':
                all_match = True
                foo = argv_word.replace(',', ' ').strip().split()
                first_word = foo[0]

                if (len(foo) >= 2 and
                    (re.match('(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}', first_word) or
                     re.match('[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}', first_word))):

                    for x in foo[1:]:
                        if '/' in x:
                            try:
                                IPNetwork(x)
                            except ValueError:
                                all_match = False
                                break
                        else:
                            all_match = False
                            break
                else:
                    all_match = False

                if all_match:
                    self.value = argv_word.replace(',', ' ').strip()
                    self.value_raw = argv_word.replace(',', ' ').strip()
                    return True

            # 4 swp1 swp2
            elif self.text == '<number-interface-list>':
                foo = argv_word.replace(',', ' ').strip().split()
                first_word = foo[0]

                if first_word.isdigit() and len(foo) >= 2:
                    self.value = argv_word.replace(',', ' ').strip()
                    self.value_raw = argv_word.replace(',', ' ').strip()
                    return True

            # swp1=yes swp2=no
            elif self.text == '<interface-yes-no-list>':

                if interface_yes_no_list(argv_word, ('yes', 'no')):
                    self.value = argv_word.replace(',', ' ').strip()
                    self.value_raw = argv_word.replace(',', ' ').strip()
                    return True

            # swp1=yes swp2=no swp3=1 swp4=0
            elif self.text == '<interface-yes-no-0-1-list>':

                if interface_yes_no_list(argv_word, ('yes', 'no', '0', '1')):
                    self.value = argv_word.replace(',', ' ').strip()
                    self.value_raw = argv_word.replace(',', ' ').strip()
                    return True

            # swp1=yes swp2=no swp3=auto
            elif self.text == '<interface-yes-no-auto-list>':

                if interface_yes_no_list(argv_word, ('yes', 'no', 'auto')):
                    self.value = argv_word.replace(',', ' ').strip()
                    self.value_raw = argv_word.replace(',', ' ').strip()
                    return True

            elif self.text.startswith('<interface'):
                if ' ' not in argv_word:
                    if not argv_word.isdigit():
                        self.value = argv_word
                        self.value_raw = argv_word
                        return True

            elif self.text == '<number>' or self.text.startswith('<number-'):
                if argv_word.isdigit():
                    self.value = int(argv_word)
                    self.value_raw = argv_word
                    return True

            elif self.text == '<text>' or self.text.startswith('<text-'):
                if ' ' not in argv_word:
                    self.value = argv_word
                    self.value_raw = argv_word
                    return True

            else:
                re_number_range = re.match('<(\d+)-(\d+)>', self.text)

                if re_number_range:

                    if argv_word.isdigit():
                        min_int = int(re_number_range.group(1))
                        max_int = int(re_number_range.group(2))
                        argv_word = int(argv_word)

                        if argv_word >= min_int and argv_word <= max_int:
                            self.value = argv_word
                            self.value_raw = argv_word
                            return True

                        # The user entered a number it is just out of range
                        else:
                            if use_fuzzy:
                                self.value = argv_word
                                self.value_raw = argv_word
                                self.fuzzy = True
                                return True
                else:
                    raise Exception("Add support for matching %s" % self.text)

        elif self.text.isdigit():
            if argv_word.isdigit() and int(self.text) == int(argv_word):
                self.value = int(argv_word)
                self.value_raw = int(argv_word)
                return True

        else:
            if self.text == argv_word:
                self.value = argv_word
                self.value_raw = argv_word
                return True

            if self.len_text > len(argv_word):
                if self.text.startswith(argv_word):
                    self.value = self.text
                    self.value_raw = argv_word
                    return True

            if use_fuzzy:
                # commenting this out for now...it tends to create more head scratching scenarios than it does give goo
                # if self.soundex == soundex(argv_word):
                #     log.debug("FUZZY MATCH (soundex): %s vs %s" % (self.text, argv_word))
                #     self.value = self.text
                #     self.value_raw = argv_word
                #     self.fuzzy = True
                #     return True

                for typoed_word in swap_letter_pairs(self.text):
                    if typoed_word == argv_word:
                        log.debug("FUZZY MATCH (letter swap typo): '%s' should have been '%s'" % (typoed_word, self.text))
                        self.value = self.text
                        self.value_raw = argv_word
                        self.fuzzy = True
                        return True

                for typoed_word in drop_each_letter(self.text):
                    if typoed_word == argv_word:
                        log.debug("FUZZY MATCH (letter drop typo): '%s' should have been '%s'" % (typoed_word, self.text))
                        self.value = self.text
                        self.value_raw = argv_word
                        self.fuzzy = True
                        return True

        return False


class Node(object):

    def __init__(self, text):
        self.text = text
        self.value = None  # will store the Word object that matched
        self.words = []
        self.ambiguity_score = 0

        # Node is required?
        if self.text.startswith('['):
            self.required = False
            self.text = self.text[1:-1]

        elif self.text.startswith('('):
            assert '|' in self.text, "Nodes with () must have multiple options such as (foo|bar), this node has %s" % self.text
            self.required = True
            self.text = self.text[1:-1]

        else:
            self.required = True

        if '|' in text:
            for word_text in self.text.split('|'):
                self.words.append(Word(word_text))
        else:
            self.words.append(Word(self.text))

    def __str__(self):
        output = []
        for word in self.words:
            output.append(str(word))

        if self.required:
            if len(self.words) > 1:
                return '(' + '|'.join(output) + ')'
            else:
                return ''.join(output)
        else:
            return '[' + '|'.join(output) + ']'

    def reset(self):
        self.value = None

        for word in self.words:
            word.reset()

    def match(self, argv_word, use_fuzzy, keywords_only, tree_options):
        """
        Save the Word object that matched as self.value
        """
        matches = []
        for word in self.words:
            if keywords_only:
                if not word.text.startswith('<') and word.match(argv_word, use_fuzzy, tree_options):
                    matches.append(word)
            else:
                if word.match(argv_word, use_fuzzy, tree_options):
                    matches.append(word)

        # Say you have two options, <ip> and <interface>. <interface> will match
        # on anything without a whitespace so 1.1.1.1 will match both <ip> and <interface>
        if len(matches) > 1:
            var_strict_matches = []
            var_loose_matches = []
            keyword_matches = []

            # Find how many of the matches were against a variable like <ip>, <text>, etc
            for word in matches:
                if word.text.startswith('<'):
                    re_number_range = re.match('<(\d+)-(\d+)>', self.text)

                    if (re_number_range or
                        word.text in ('<ip>', '<ip/prefixlen>', '<ipv4>', '<ipv4/prefixlen>', '<ipv6>', '<ipv6/prefixlen>', '<number>', '<mac>') or
                        word.text.startswith('<ip-') or
                        word.text.startswith('<ipv4-') or
                        word.text.startswith('<ipv6-') or
                        word.text.startswith('<number-')):
                        var_strict_matches.append(word)

                    elif is_loose_matching_variable(word.text):
                        var_loose_matches.append(word)

                    else:
                        raise Exception("Do not know if %s is a loose or strict match" % word.text)
                else:
                    keyword_matches.append(word)

            # These are all matching the same basic thing so for now pick the first one
            if len(var_loose_matches) > 1:
                var_loose_matches = [var_loose_matches[0], ]

            # log.info("var_strict_matches %s" % map(str, var_strict_matches))
            # log.info("var_loose_matches %s" % map(str, var_loose_matches))
            # log.info("keyword_matches %s" % map(str, keyword_matches))

            # If we have either of these then ignore the loose matches
            if var_strict_matches or keyword_matches:
                matches = var_strict_matches + keyword_matches
            elif var_loose_matches:
                matches = var_loose_matches

        self.matches = []

        # There was only one Word that matched so this is easy
        if len(matches) == 1:
            self.value = matches[0]
            self.matches = matches
            return True

        elif matches:

            # If a node has choices (upstream|upstream-join-desired|upstream-rpf)
            # and the user enters 'upstream' we will have three matches...we should
            # - find the shortest word among all choices
            # - return False if any of the matches is a var like <ipv4>, etc
            # - return True if all of our choices start with the shortest word
            all_match_shortest_word = True
            shortest_word_len = None
            shortest_word = None

            for word in matches:
                if word.text.startswith('<'):
                    all_match_shortest_word = False
                    break

                word_len = len(word.value)
                if shortest_word_len is None or word_len < shortest_word_len:
                    shortest_word_len = word_len
                    shortest_word = word

            if all_match_shortest_word:
                for word in matches:
                    # print word
                    if not word.value.startswith(shortest_word.value):
                        all_match_shortest_word = False
                        break

            # log.info("all_match_shortest_word %s" % all_match_shortest_word)
            # log.info("shortest_word_len %s" % shortest_word_len)
            # log.info("shortest_word %s" % shortest_word)

            if all_match_shortest_word:
                self.value = shortest_word
                self.matches = matches
                return True

            # All start with same text
            # Only do this if we are being called via tree_options()...not tree_match()
            if tree_options:
                all_start_with_same_text = True

                for word in matches:
                    if word.text.startswith('<') or not word.text.startswith(argv_word):
                        all_start_with_same_text = False
                        break

                # log.info("all_start_with_same_text %s" % all_start_with_same_text)
                if all_start_with_same_text:
                    self.value = shortest_word
                    self.matches = matches
                    return True

        # Nothing matched
        return False


def get_indexed_nodes(text):
    """
    Most of the time whitespace can be used to split nodes
        (set|clear) <interface> clagd-enable (no|yes)

    nodes
    - (set|clear)
    - <interface>
    - clagd-enable
    - (no|yes)

    But if we are dealing with multiword keywords, such as "soft in", that approach
    does not work. We can only split on whitespaces if we are not inside a () or []
        bgp (<ipv4>|<ipv6>|<interface>|*) [soft in|soft out]

    nodes:
    - bgp
    - (<ipv4>|<ipv6>|<interface>|*)
    - [soft in|soft out]
    """
    nodes = []
    node_index = 0
    node_text = []
    parens = 0
    brackets = 0

    for char in text:
        if char == ' ':
            if parens == 0 and brackets == 0:
                nodes.append((node_index, ''.join(node_text)))
                node_index += 1
                node_text = []
            else:
                node_text.append(char)
        else:
            if char == '(':
                parens += 1

            elif char == ')':
                parens -= 1

            elif char == '[':
                brackets += 1

            elif char == ']':
                brackets -= 1

            node_text.append(char)

    if node_text:
        nodes.append((node_index, ''.join(node_text)))

    return nodes


class Command(object):

    def __init__(self, text):
        assert text.count('[') == text.count(']'), 'You have mis-matched []s in:\n  %s\n' % text
        assert text.count('(') == text.count(')'), 'You have mis-matched ()s in:\n  %s\n' % text
        self.text = text.strip()
        self.nodes = []
        self.node_strings = []

        for (_, node_text) in get_indexed_nodes(self.text):
            self.node_strings.append(node_text)

            # These are only used by get_keyword_help()
            # The Node objects that are used 99% of the time live in the 'tree'
            self.nodes.append(Node(node_text))

        self.program = [self.node_strings[0]]

        # log a warning if a command uses the same <> multiple times.  For example
        # here <0-32> is used twice, we'll send up storing the values in self.args['<0-32>']
        # so we cannot have more than one of these since they would collide in self.args (it is a dictionary)
        #
        # net (add|del) ip prefix-list <text> (deny|permit) <ipv4/prefixlen> ge <0-32> [le <0-32>]
        node_string_words = []

        for node in self.nodes:
            for word in node.words:
                for x in word.text.split():
                    if x.startswith('<') and x in node_string_words:
                        log.warning("%s appears multiple times, this is not supported\n%s\n" % (x, self.text))
                    node_string_words.append(x)

    def __str__(self):
        return ' '.join(self.node_strings)


def tree_words(tree):
    """
    Recursively walk the tree and return all words in the tree
    """
    words = []

    # node is a Node object
    # leaves is an OrderedDict of other Node objects
    for node, leaves in tree.items():
        for word in node.words:
            words.append(word.text)

        if leaves:
            words.extend(tree_words(leaves))

    return list(set(words))


def tree_reset_nodes(tree):
    """
    Recursively walk the tree and reset each node
    """
    # node is a Node object
    # leaves is an OrderedDict of other Node objects
    for node, leaves in tree.items():
        node.reset()

        if leaves:
            tree_reset_nodes(leaves)


def tree_match(tree, argv, use_fuzzy):
    """
    Recursively walk the tree and return all branches that are matched by argv
    """
    matches = {}

    # node_strings = []
    # for node, leaves in tree.items():
    #     node_strings.append(str(node))
    # log.debug("tree_match() look for '%s' in nodes %s " % (None if not argv else argv[0], str(node_strings)))

    # node is a Node object
    # leaves is an OrderedDict of other Node objects
    for node, leaves in tree.items():
        # log.debug("%s: tree_match leaves %s, remaining argv '%s'" % (node, map(str, leaves), ' '.join(argv)))

        if argv and node.match(argv[0], use_fuzzy, keywords_only=False, tree_options=False):
            argv_matched =  argv[0]

            # Make a copy and chop the leading word that was used to match this node
            newargv = deepcopy(argv)

            # If this is a wildcard overwrite the Word.value to hold everything left in argv
            if node.value.text in ('<wildcard>', '<regex>') or node.text.startswith('<wildcard-'):
                node.value.value = ' '.join(argv)
                newargv = []
            else:
                newargv = newargv[1:]

            log.debug("tree_match() MATCH - node %s vs argv %s, leaves %s, remaining argv '%s'" %
                      (node, argv_matched, map(str, leaves), ' '.join(newargv)))

            if leaves:
                ret = tree_match(leaves, newargv, use_fuzzy)

                # If none of our leaves matched the remaining argv then this
                # is a dead end so continue on to the next node
                if ret is None:
                    continue
                else:
                    matches[node] = ret

            else:
                # We are at the end of the branch, if there are still argv left
                # to match then this branch was not a match
                if newargv:
                    continue

                # There are no leaves left but no argv either so this is a match
                else:
                    matches[node] = {}

        # We didn't match but this node is optional
        elif not node.required:
            log.debug("tree_match() NO MATCH (optional) - node %s vs argv %s, leaves %s" % (node, None if not argv else argv[0], map(str, leaves)))

            if leaves:
                ret = tree_match(leaves, argv, use_fuzzy)

                if ret is None:
                    continue
                else:
                    matches[node] = ret

            else:
                # We are at the end of the branch where the last node is optional
                # If there is anything left in argv then this branch did not match
                if argv:
                    continue
                else:
                    matches[node] = {}
        else:
            # This is a bit chatty so commenting out
            # log.debug("tree_match() NO MATCH (required) - node %s vs argv %s, leaves %s" % (node, argv[0], map(str, leaves)))
            pass

    if matches:
        log.debug("tree_match() node %s (%s) returning %s" % (node, "required" if node.required else "optional", map(str, matches)))
        return matches

    # If we are here it means this is a required Node that did not match
    else:
        log.debug("tree_match() node %s (%s) returning None" % (node, "required" if node.required else "optional"))
        return None


def tree_options(tree, argv):
    """
    Recursively walk the tree to find TAB complete options
    """
    matches = {}

    log.debug("")
    node_strings = []
    for node, leaves in tree.items():
        node_strings.append(str(node))
    log.debug("tree_options() look for '%s' in nodes %s " % (None if not argv else argv[0], str(node_strings)))

    # node is a Node object
    # leaves is an OrderedDict of other Node objects
    for node, leaves in tree.items():

        if argv and node.match(argv[0], use_fuzzy=False, keywords_only=False, tree_options=True):
            argv_matched = argv[0]

            # Make a copy and chop the leading word that was used to match this node
            newargv = deepcopy(argv)

            # If this is a wildcard overwrite the Word.value to hold everything left in argv
            if node.value.text in ('<wildcard>', '<regex>') or node.text.startswith('<wildcard-'):
                node.value.value = ' '.join(argv)
                newargv = []
            else:
                newargv = newargv[1:]

            log.debug("tree_options() MATCH - node %s vs argv %s, leaves %s, remaining argv '%s'" %
                      (node, argv_matched, map(str, leaves), ' '.join(newargv)))

            # This check is the primary difference between tree_match vs tree_options
            if leaves and newargv:
                ret = tree_options(leaves, newargv)

                if ret is None:
                    continue
                else:
                    matches[node] = ret

            else:
                # We are at the end of the branch, if there are still argv left
                # to match then this branch was not a match
                if newargv:
                    continue

                # There are no leaves left but no argv either so this is a match
                else:
                    matches[node] = {}

        # We didn't match but this node is optional
        elif not node.required:
            log.debug("tree_options() NO MATCH (optional) - node %s vs argv %s, leaves %s" % (node, None if not argv else argv[0], map(str, leaves)))

            if leaves:
                ret = tree_options(leaves, argv)

                if ret is None:
                    continue
                else:
                    matches[node] = ret

            else:
                # We are at the end of the branch where the last node is optional
                # If there is anything left in argv then this branch did not match
                if argv:
                    continue
                else:
                    matches[node] = {}
        else:
            # This is a bit chatty so commenting out
            # log.debug("tree_options() NO MATCH (required) - node %s vs argv %s, leaves %s" % (node, argv[0], map(str, leaves)))
            pass

    if matches:
        log.debug("tree_options() tree %s returning %s" % (str(node_strings), map(str, matches)))
        return matches

    # If we are here it means this is a required Node that did not match
    else:
        log.debug("tree_options() tree %s returning None" % str(node_strings))
        return None


def tree_option_words(tree):
    """
    Recursively walk the tree and return all words from the child nodes
    """
    words = []

    for node, leaves in tree.items():
        words.extend(node.words)

        if not node.required and leaves:
            words.extend(tree_option_words(leaves))

    return words


def tree_find_node(tree, target_node):

    for node, leaves in tree.items():
        if node == target_node:
            return leaves
        else:
            ret = tree_find_node(leaves, target_node)
            if ret:
                return ret
    return None


def tree_has_one_branch(tree):
    """
    Recursively walk the tree and return True if there is only one branch
    """
    for node, leaves in tree.items():
        if leaves:
            if len(leaves.keys()) > 1:
                return False
            else:
                return tree_has_one_branch(leaves)
    return True


def tree_populate_args(tree, args):
    matches = []

    for node, leaves in tree.items():
        matches.append(node)
        matched_word = node.value

        # If the node was optional this could be None
        if matched_word:
            args[matched_word.text] = matched_word.value
            # print "%s: matched_word %s, value %s" % (node, matched_word.text, matched_word.value)

        if leaves:
            matches.extend(tree_populate_args(leaves, args))

    return matches


def tree_leaf_nodes(tree):
    """
    Recursively walk the tree and return all of the leaf Nodes
    """
    leaf_nodes = []

    for node, leaves in tree.items():
        if leaves:
            leaf_nodes.extend(tree_leaf_nodes(leaves))
        else:
            leaf_nodes.append(node)

    return leaf_nodes


def tree_branches_strings(tree):
    """
    Returns a generator that will produce a list of the branches in a tree
    For example if your tree is:

    foo
     +-- clear
     |   +-- ip
     |       +-- bgp
     |           +-- *
     +-- show
         +-- ip
             +-- bgp
             |   +-- [ipv4 unicast|ipv4 multicast|ipv6 unicast]
             |   |   +-- summary
             |   +-- neighbor
             |       +-- [<ip>]
             +-- route

    The returned generator will contain a list of the following lines:
        clear ip bgp *
        show ip bgp [ipv4 unicast|ipv4 multicast|ipv6 unicast] summary
        show ip bgp neighbor [<ip>]
        show ip route
    """
    for node, leaves in tree.items():

        if leaves:
            for text in tree_branches_strings(leaves):
                # If it was a fuzzy match print it in blue so the user can see which word they typoed
                if node.value and node.value.fuzzy:
                    yield '\033[%dm%s\033[0m %s' % (blue, node, text)
                else:
                    yield '%s %s' % (node, text)
        else:
            # If it was a fuzzy match print it in blue so the user can see which word they typoed
            if node.value and node.value.fuzzy:
                yield "\033[%dm%s\033[0m" % (blue, node)
            else:
                yield str(node)


def tree_find_branch(tree, target_leaf_node):
    """
    Given a leaf node, find the branch that leads to that leaf node
    """
    matches = {}

    for node, leaves in tree.items():
        if leaves:
            ret = tree_find_branch(leaves, target_leaf_node)

            if ret:
                matches[node] = ret

        else:
            if node == target_leaf_node:
                matches[node] = {}

    if matches:
        return matches
    else:
        return None


def charcacter_diff_count(stringA, stringB):
    # TODO this should account for more than length
    return abs(len(stringA) - len(stringB))


def tree_calc_ambiguity_scores(tree, score=0):
    """
    Recursively walk the tree and calculate the ambiguity score for
    the branch, store the score in the leaf node
    """

    for node, leaves in tree.items():
        if node.value:
            if node.value.text.startswith('<'):
                pass
            else:
                delta = charcacter_diff_count(node.value.value, node.value.value_raw)
                if delta:
                    score += delta
                    log.debug("%s ambiguity score is now %d, delta %d (%s vs %s)" % (node, score, delta, node.value.value, node.value.value_raw))

        if leaves:
            tree_calc_ambiguity_scores(leaves, score)
        else:
            node.ambiguity_score = score

    return score


def tree_least_ambiguous(tree):
    tree_calc_ambiguity_scores(tree)
    min_ambiguity_score = None
    min_ambiguity_score_leaf_node = None

    for leaf_node in tree_leaf_nodes(tree):
        log.debug("%s ambiguity score %d" % (leaf_node, leaf_node.ambiguity_score))
        if min_ambiguity_score is None or leaf_node.ambiguity_score < min_ambiguity_score:
            min_ambiguity_score = leaf_node.ambiguity_score
            min_ambiguity_score_leaf_node = leaf_node

    log.debug("%s with ambiguity score %s is the least ambiguous" % (min_ambiguity_score_leaf_node, min_ambiguity_score))
    return tree_find_branch(tree, min_ambiguity_score_leaf_node)


def format_help_text(text, width):
    """
    Take a really long string of help text and break it down into multiple
    lines all less than or equal to 'width'
    """
    lines = []
    line = []
    line_length = 0

    if not text:
        text = 'help text would go here'
        #return lines

    for word in text.split():

        # Too long
        if (line_length + len(word)) > width:
            lines.append(' '.join(line))
            line = []
            line_length = 0

        line.append(word)
        line_length += len(word) + 1  # +1 for the whitespace after the word

    if line:
        lines.append(' '.join(line))

    return lines


def expand_nodes_to_lines(nodes):
    lines = []
    line = []

    for (node_index, node) in enumerate(nodes):
        if isinstance(node, str):
            line.append(node)
        else:
            # 'node' is a list of strings
            for x in node:
                for rest_of_line in expand_nodes_to_lines(nodes[node_index+1:]):
                    # compress duplicate whitespaces
                    lines.append(' '.join(line + [x, ] + [rest_of_line, ]).strip().replace('  ', ' '))
            break
    else:
        line = ' '.join(line)
        lines.append(line)

    return lines


def expand_line_to_multiple_commands(line):
    added_line = False
    program = None
    nodes = []

    for (node_index, node_text) in get_indexed_nodes(line):
        if not program:
            program = node_text

        if ' ' in node_text:
            # Originally I tried putting nodes on the tree that contained multiple
            # keywords ('ipv4 unicast')...this made TAB completion a nightmare though.
            # It is much cleaner to expand these options out and build a bigger tree.

            # For 'show ip bgp [ipv4 unicast|ipv4 multicast|ipv6 unicast] summary'
            # Expand this out to the four lines that will cover each scenario
            #
            # show ip bgp ipv4 multicast summary
            # show ip bgp ipv4 unicast summary
            # show ip bgp ipv6 unicast summary
            # show ip bgp summary
            if node_text.startswith('['):
                nodes.append(node_text[1:-1].split('|') + [''])

            # For 'show ip bgp (ipv4 unicast|ipv4 multicast|ipv6 unicast) summary'
            # Expand this out to the three lines that will cover each scenario
            #
            # show ip bgp ipv4 multicast summary
            # show ip bgp ipv4 unicast summary
            # show ip bgp ipv6 unicast summary
            elif node_text.startswith('('):
                nodes.append(node_text[1:-1].split('|'))

            else:
                raise Exception("'%s' has a space but no [] or ()" % node_text)

        # No spaces in the node text, just add it as is
        else:
            nodes.append(node_text)

    lines = expand_nodes_to_lines(nodes)

    return (program, lines)


class NetworkDocopt():
    """
    Heavily influenced by docopt but designed to be a little more like
    a Networking CLI with partial word acceptance, IPv4 sanity checking, etc
    """

    def __init__(self, docstring, daemon=False):
        self.args = {}
        self.docstring = docstring
        self.commands = []
        self.tab_complete_hook = {}
        self.tab_complete_policer = None

        # This is only neede for dumping the tree in debugs
        try:
            from asciitree import LeftAligned
            self.tree_printer = LeftAligned()
        except ImportError:
            self.tree_printer = None

        self.match = None
        self.argv_expanded = []
        self.helptext = {}
        self.program = None
        self.daemon = daemon

        state = None
        found_help_option = False
        usage_commands = []
        self.interfaces = sort_for_humans([x for x in os.listdir('/sys/class/net/') if x != 'bonding_masters'])

        # Create a Command object for each line in the Usage section. We store
        # the lines in usage_commands so we can sort them...just to make
        # troubleshooting easier later
        for line in self.docstring.splitlines():

            if line.startswith('Usage'):
                state = 'USAGE'
                continue

            elif line.startswith('Help'):
                state = 'HELP'

            elif line.startswith('Options'):
                state = 'HELP'

            # Ignore comments in the docstring
            elif line.lstrip().startswith('#'):
                continue

            elif line == '':
                continue

            if state == 'USAGE':
                if 'help' in line:
                    found_help_option = True

                # Compress multiple spaces down to a single space
                line = ' '.join(line.strip().split())

                # Remove spaces that are not needed, they will cause problems later
                line = line.replace(' | ', '|')
                line = line.replace('| ', '|')
                line = line.replace(' |', '|')
                line = line.replace('( ', '(')
                line = line.replace(' )', ')')
                line = line.replace('[ ', '[')
                line = line.replace(' ]', ']')

                (self.program, lines) = expand_line_to_multiple_commands(line)
                usage_commands.extend(lines)

            elif state == 'HELP':
                line = line.strip()
                re_line = re.search('^(.*?):(.*)$', line)

                if re_line:
                    keyword = re_line.group(1).strip()
                    keyword_help = re_line.group(2).strip()
                    if keyword and keyword_help:
                        self.helptext[keyword] = keyword_help

        for line in sorted(list(set(usage_commands))):
            self.commands.append(Command(line))

        if not found_help_option:
            line = '%s help' % self.program
            self.commands.append(Command(line))

        # Build a parse tree based on all of our Command objects
        self.tree = OrderedDict()
        self.first_node = None

        log.debug("\n\n__doc__\n=======\n%s\n" % self.docstring)
        log.debug("\n\nCOMMANDS\n========\n%s\n" % '\n'.join(map(str, self.commands)))
        log.debug("\n\nHELP TEXT\n========\n%s\n" % pformat(self.helptext))
        prev_command = None

        for command in self.commands:
            tree_node = self.tree

            if prev_command is not None:
                assert prev_command.program == command.program,\
                    "program name mismatch %s vs %s for %s" % (prev_command.program, command.program, str(command.node_strings))

            len_node_strings = len(command.node_strings)

            for (index, target_node_text) in enumerate(command.node_strings):

                if index < len_node_strings - 1:
                    next_node = command.node_strings[index+1]
                else:
                    next_node = None

                # log.debug("target node %s, next node %s" % (target_node_text, next_node))

                # Look to see if this node is already in the tree
                for (node, leaves) in tree_node.iteritems():
                    if str(node) == target_node_text:

                        if index == 0:
                            tree_node = tree_node[node]
                            break

                        # If this is not the first node then we must compare the name of the
                        # node and the leaves hanging off that node before we can declare 'node'
                        # a match for our target node
                        else:
                            leaves_str = map(str, leaves)

                            if leaves_str:
                                leaves_str = leaves_str[0]
                            else:
                                leaves_str = None

                            # log.debug("matched node %s, leaves %s" % (node, leaves_str))

                            if next_node is None:
                                if not leaves:
                                    # log.debug("None leaves do match")
                                    tree_node = tree_node[node]
                                    break
                                # else:
                                #     log.debug("None leaves do match, None vs %s" % leaves_str)

                            else:
                                if next_node == leaves_str:
                                    # log.debug("leaves do match")
                                    tree_node = tree_node[node]
                                    break
                                # else:
                                #     log.debug("leaves do NOT match %s vs %s" % (next_node, leaves_str))

                # Add this node to the tree
                else:
                    node = Node(target_node_text)
                    tree_node[node] = OrderedDict()
                    tree_node = tree_node[node]

                    if not self.first_node:
                        self.first_node = node

            # log.debug('\n')
            prev_command = command

        # print the tree
        log.debug("\n\nTREE\n====\n%s\n" % self)

        # init self.args, add a key/value for every word in the tree
        for word in sorted(tree_words(self.tree)):
            self.args[word] = None

        log.debug("\n\nWORDS\n=======\n%s\n" % pformat(self.args))

        branches = list(tree_branches_strings(self.tree[self.first_node]))
        log.debug("\n\nBRANCHES\n========\n%s\n" % '\n'.join(branches))

    def __str__(self):
        return self.get_ascii_tree(self.tree)

    def get_ascii_tree(self, tree):
        if self.tree_printer:
            return self.tree_printer(tree)
        else:
            return ''

    def re_init(self):
        self.match = None

        # init self.args, add a key/value for every word in the tree
        for word in sorted(tree_words(self.tree)):
            self.args[word] = None

        tree_reset_nodes(self.tree)

    def add_tag_matcher(self, keyword, matcher, is_loose_matcher):
        global custom_matchers
        global loose_matching_variables
        custom_matchers[keyword] = matcher

        if is_loose_matcher:
            loose_matching_variables.append(keyword)

    def del_tag_matcher(self, keyword):
        global custom_matchers
        global loose_matching_variables

        if keyword in custom_matchers:
            del custom_matchers[keyword]

        if keyword in loose_matching_variables:
            loose_matching_variables.remove(keyword)

    def add_tab_complete_hook(self, keyword, completer):
        self.tab_complete_hook[keyword] = completer

    def del_tab_complete_hook(self, keyword):
        if keyword in self.tab_complete_hook:
            del self.tab_complete_hook[keyword]

    def get(self, keyword):
        return self.args.get(keyword)

    def get_keyword_help(self, keywords):

        # This should not happend
        if not keywords:
            return 'ERROR: get_keyword_help() was given an empty keyword list'

        commands_with_keywords = []

        for command in self.commands:
            tmp_keywords = deepcopy(keywords)

            for node in command.nodes:
                match = None

                for keyword in tmp_keywords:
                    if node.match(keyword, use_fuzzy=False, keywords_only=True, tree_options=True):
                        match = keyword
                        break

                if match:
                    tmp_keywords.remove(match)

            # If we matched all of the keywords then the keywords list will be empty
            if not tmp_keywords:
                commands_with_keywords.append(str(command))

        if commands_with_keywords:
            output = ['', "The following commands contain keyword(s) '%s':" % "', '".join(keywords), '']
            for command in sorted(commands_with_keywords):
                output.append('    ' + command)
            output.append('')
            output.append('')
            return '\n'.join(output)
        else:
            return "ERROR: There are no commands with keyword(s) '%s'" % "', '".join(keywords)

    def run(self, argv=None):
        self.argv_expanded = []
        self.match = None

        # The 1st item in argv is the program name...ignore it
        if argv:
            self.argv = argv
        else:
            self.argv = sys.argv

        # argv[0] will be the program name, if the program name is foo but the
        # user ran it as ./foo remove everything before the /
        if '/' in self.argv[0]:
            self.argv[0] = self.argv[0].split('/')[-1]

        log.debug("ARGV (run): %s" % self.argv)
        len_argv = len(self.argv)

        # If the user did not provide any options there is nothing we can do
        # other than print the docstring
        if len_argv == 1:
            print self.docstring
            return

        help_tuple = ('-h', '--help', 'help', 'hel', 'he', 'h')

        # net help
        if len_argv == 2 and self.argv[1] in help_tuple:
            print self.docstring
            self.args['help'] = True

        # net help KEYWORDS
        elif self.argv[1] in help_tuple:
            print self.get_keyword_help(self.argv[2:])
            self.args['help'] = True

        # net KEYWORDS help
        elif self.argv[-1] in help_tuple:
            print self.get_keyword_help(self.argv[1:-1])
            self.args['help'] = True

        else:
            matches = tree_match(self.tree, self.argv, use_fuzzy=False)

            if matches:

                if tree_has_one_branch(matches):
                    log.debug("\n\nMATCHES (tree)\n==============\n%s" % self.get_ascii_tree(matches))
                    self.match = tree_populate_args(matches, self.args)
                    log.debug("\n\nMATCHES (list)\n==============\n%s" % map(str, self.match))

                else:
                    log.debug("\n\nMATCHES AMBIGUOUS (tree)\n==============\n%s" % self.get_ascii_tree(matches))
                    matches = tree_least_ambiguous(matches)
                    log.debug("\n\nMATCHES (tree)\n==============\n%s" % self.get_ascii_tree(matches))

                    if tree_has_one_branch(matches):
                        self.match = tree_populate_args(matches, self.args)
                        log.debug("\n\nMATCHES (list)\n==============\n%s" % map(str, self.match))
                    else:
                        raise Exception("TODO: pick the branch that is the least ambiguous")

                if self.match:
                    for node in self.match:
                        if node.value:
                            self.argv_expanded.append(node.value.value)
                    log.debug("\n\nMATCHES (argv_expanded)\n==============\n%s" % pformat(self.argv_expanded))

            else:
                log.debug('\n\n\n\n')
                log.debug("DID NOT MATCH - find did you mean matches")
                did_you_mean_tree = tree_match(self.tree, self.argv, use_fuzzy=True)
                print "ERROR: Command not found\n"

                if did_you_mean_tree:
                    log.debug("DID YOU MEAN (tree):\n%s" % self.get_ascii_tree(did_you_mean_tree))

                    blue_aspathlist_string = '\033[%dm<as-path-access-list>' % blue
                    blue_aspathlist_optional_string = '\033[%dm[<as-path-access-list>' % blue
                    blue_commlist_string = '\033[%dm<community-list>' % blue
                    blue_commlist_optional_string = '\033[%dm[<community-list>' % blue
                    blue_extcommlist_string = '\033[%dm<extcommunity-list>' % blue
                    blue_extcommlist_optional_string = '\033[%dm[<extcommunity-list>' % blue
                    blue_bgppeer_string = '\033[%dm<bgppeer>' % blue
                    blue_ip_string = '\033[%dm<ip>' % blue
                    blue_ipv4_string = '\033[%dm<ipv4>' % blue
                    blue_ipv6_string = '\033[%dm<ipv6>' % blue
                    blue_ip_prefixlen_string = '\033[%dm<ip/prefixlen>' % blue
                    blue_ipv4_prefixlen_string = '\033[%dm<ipv4/prefixlen>' % blue
                    blue_ipv6_prefixlen_string = '\033[%dm<ipv6/prefixlen>' % blue
                    blue_prefixlist_string = '\033[%dm<prefix-list>' % blue
                    blue_prefixlist_optional_string = '\033[%dm[<prefix-list>' % blue
                    blue_routemap_string = '\033[%dm<route-map>' % blue
                    blue_routemap_optional_string = '\033[%dm[<route-map>' % blue

                    print 'Did you mean one of the following?\n'
                    for branch in tree_branches_strings(did_you_mean_tree):
                        print '    ' + branch
                        print_newline = False

                        if blue_ip_string in branch or blue_ipv4_string in branch or blue_ipv6_string in branch:
                            print "        This command is looking for an IP address, not an IP/prefixlen"
                            print_newline = True

                        if blue_ip_prefixlen_string in branch or blue_ipv4_prefixlen_string in branch or blue_ipv6_prefixlen_string in branch:
                            print "        This command is looking for an IP/prefixlen, not an IP address"
                            print_newline = True

                        if blue_bgppeer_string in branch:
                            print "        This command is looking for an existing BGP peer"
                            print_newline = True

                        if blue_aspathlist_string in branch or blue_aspathlist_optional_string in branch:
                            print "        This command is looking for an existing as-path access-list"
                            print_newline = True

                        if blue_commlist_string in branch or blue_commlist_optional_string in branch:
                            print "        This command is looking for an existing community-list"
                            print_newline = True

                        if blue_extcommlist_string in branch or blue_extcommlist_optional_string in branch:
                            print "        This command is looking for an existing extcommunity-list"
                            print_newline = True

                        if blue_prefixlist_string in branch or blue_prefixlist_optional_string in branch:
                            print "        This command is looking for an existing prefix-list"
                            print_newline = True

                        if blue_routemap_string in branch or blue_routemap_optional_string in branch:
                            print "        This command is looking for an existing route-map"
                            print_newline = True

                        if print_newline:
                            print ''

                # If we got here then we have no freaking clue what command
                # the user was trying to run
                else:
                    print "Use '%s help KEYWORD(s)' to list all options that use KEYWORD(s)" % self.program

    def scrub_loose_options(self, final_words, found_optional_branch, ended_with_space, last_argv):
        """
        # This scenario
        prompt# net clear ip bgp pref --completions 4
    <interface>  :  An interface name
    prefix       :  add help text
    <ENTER>

        # needs to return just 'prefix' so that it can tab complete
        prompt# net clear ip bgp pref --completions 4
    prefix       :  add help text
        """
        if ended_with_space:
            return (final_words, found_optional_branch)

        all_vars_are_loose = True
        all_keywords_start_with_last_argv = True
        start_with_last_argv = []

        for word in final_words:
            if word.startswith('<'):
                if not is_loose_matching_variable(word):
                    all_vars_are_loose = False
                    break
            else:
                if word.startswith(last_argv):
                    start_with_last_argv.append(word)
                else:
                    all_keywords_start_with_last_argv = False
                    break

        if all_vars_are_loose and all_keywords_start_with_last_argv:
            return (start_with_last_argv, False)

        return (final_words, found_optional_branch)

    def print_options(self, ended_with_space, options_style='bash', argv=None):
        last_argv = None
        found_optional_branch = False

        if argv:
            self.argv = argv
        else:
            self.argv = sys.argv

        # argv[0] will be the program name, if the program name is foo but the
        # user ran it as ./foo remove everything before the /
        if '/' in self.argv[0]:
            self.argv[0] = self.argv[0].split('/')[-1]

        log.debug("ARGV (ended_with_space %s): %s" % (ended_with_space, self.argv))
        option_words = []

        if len(self.argv) > 1:
            last_argv = self.argv[-1]
            options = tree_options(self.tree, self.argv)

            if options:
                # log.debug("OPTIONS (dict):\n%s" % map(str, options))
                log.debug("OPTIONS (tree):\n%s" % self.get_ascii_tree(options))
                leaf_nodes = tree_leaf_nodes(options)

                # Even though we are only looking for tab complete options we still
                # populate self.args based on the options tree...we do this so any
                # tab_complete_XXX() hooks can be given self.args.  The hooks may need
                # to know what args matched in order to determine what to tab complete.
                tree_populate_args(options, self.args)

                for leaf_node in leaf_nodes:
                    leaf_node_word = leaf_node.value

                    if leaf_node_word and leaf_node_word.value_raw == last_argv:
                        last_argv_matched = True
                    else:
                        last_argv_matched = False

                    # If the last node in the branch is a <wildcard> we want to append
                    # <wildcard> as an option so that the custom tab_completer for
                    # <wildcard> fires.
                    if leaf_node.text.startswith('<wildcard'):
                        option_words.append(leaf_node)

                    log.debug("OPTIONS LEAF: %s, last_argv_matched %s (%s vs %s)" % (leaf_node, last_argv_matched, leaf_node_word.value, last_argv))
                    leaves = tree_find_node(self.tree[self.first_node], leaf_node)

                    # If our tree has "show ip bgp summary" and the user typed "show ip bgp <tab>"
                    # then we need to return "summary" as the option. If the user typed
                    # "show ip bgp<TAB>" (no space at the end) then we need to return "bgp" as the option.
                    if ended_with_space:

                        if leaves:
                            option_words.extend(tree_option_words(leaves))

                    # For '(add|del) interface <interface> ip ospf network (point-to-point|point-to-multipoint|broadcast)'
                    # - 'add int swp2 ip ospf netwo<TAB>' must return "network"
                    # - 'add int swp2 ip ospf netwo <TAB>' must return "point-to-point", "point-to-multipoint", "broadcast"
                    # - 'add int swp2 ip ospf netwo p<TAB>' must return "point-to-point", "point-to-multipoint"
                    #
                    # The third bullet is the scenario we cover here
                    elif not last_argv_matched:

                        if leaves:
                            for word in tree_option_words(leaves):
                                if word.match(last_argv, False, True):
                                    option_words.append(word)
                    else:
                        for word in leaf_node.matches:
                            option_words.append(word)

                    # If there is a branch off of the leaf node where every node is
                    # optional then we must add <ENTER> as an option
                    if not found_optional_branch and len(leaf_node.matches) == 1:

                        if leaves:
                            # Convert each branch to a string and see if each node starts with a [
                            for line in tree_branches_strings(leaves):
                                optional_branch = True

                                for (_, node_text) in get_indexed_nodes(line):
                                    if not node_text.startswith('['):
                                        optional_branch = False
                                        break

                                if optional_branch:
                                    found_optional_branch = True
                                    break

                        # If there are no leaves then we need to display <ENTER>
                        else:
                            found_optional_branch = True

                log.debug("OPTIONS (words): %s" % map(str, option_words))

            else:
                log.debug("NO OPTIONS")

        # If we are here the user hit TAB without providing any command args so
        # we start at the root of the tree and list all of the first level options
        else:
            option_words = tree_option_words(self.tree[self.first_node])
            log.debug("OPTIONS2 (words): %s" % map(str, option_words))

        # At this point we have all of the words to print as options. There
        # are some of these we want to expand on though. <interfaces> for
        # instance, we want users to be able to TAB complete the interface names.
        final_words = []

        option_words_text = []
        for word in option_words:
            option_words_text.append(word.text)
        option_words_text = list(set(option_words_text))
        log.debug("option_words_text: %s" % option_words_text)

        for word_text in option_words_text:
            if word_text in self.tab_complete_hook:
                completer = self.tab_complete_hook[word_text]

                for (keyword, keyword_help) in completer(self.args, ended_with_space, last_argv):
                    final_words.append(keyword)

                    if keyword_help:
                        self.helptext[keyword] = keyword_help

            elif word_text.startswith('<interface'):
                self.interfaces = sort_for_humans([x for x in os.listdir('/sys/class/net/') if x != 'bonding_masters'])

                if ended_with_space:
                    final_words.extend(self.interfaces)
                else:
                    for ifname in self.interfaces:
                        if ifname.startswith(last_argv):
                            final_words.append(ifname)

                final_words.append(word_text)

            else:
                final_words.append(word_text)

        final_words = sort_for_humans(list(set(final_words)))

        if self.tab_complete_policer:
            final_words = self.tab_complete_policer(self.args, final_words, ended_with_space, last_argv)

        (final_words, found_optional_branch) = self.scrub_loose_options(final_words, found_optional_branch, ended_with_space, last_argv)

        log.debug("OPTIONS (final): %s" % str(final_words))

        # For fish we always print options (with help text) to STDOUT
        if options_style == 'fish':
            for word in final_words:
                if word in self.helptext:
                    print "%s\t%s" % (word, self.helptext[word])
                else:
                    print word

            if found_optional_branch:
                print '<ENTER>'

        # For bash we do one of two things:
        # - print options without help text to STDOUT
        # - print options with help text but to STDOUT but flag for STDERR
        elif options_style == 'bash':
            use_stdout = False

            if last_argv:
                all_start_with_last_argv = True

                for word in final_words:
                    if not word.startswith(last_argv):
                        all_start_with_last_argv = False
                        break

                if all_start_with_last_argv:
                    '''
                    Now check to see if this would actually make us tab complete more
                    If the user does "mst<TAB>" when our options are 'mstpctl-bpduguard'
                    and 'mstpctl-portautoedge' then we want to print these to stdout.
                    But if the user does "mstpctl-<TAB>" we do not want to print these because
                    we have already TABed out as far as we can go so it would be better to
                    print the help text.
                    '''
                    first_characters = []
                    for word in final_words:
                        tmp_word = word.replace(last_argv, '')
                        if tmp_word:
                            first_characters.append(tmp_word[0])
                        else:
                            first_characters.append(None)

                    first_characters = list(set(first_characters))

                    if len(first_characters) == 1:
                        use_stdout = True

                    # log.info("last_argv: %s" % last_argv)
                    # log.info("final_words: %s" % final_words)
                    # log.info("first_characters %s" % first_characters)

            if use_stdout:
                for word in final_words:
                    print word

            else:
                output = []

                # If we are running as a daemon the daemon should catch everything we print
                # so write to stdout but use the USE_STDERR output to signal to the daemon
                # that it should print this to stderr.
                if self.daemon:
                    output.append("USE_STDERR")
                else:
                    output.append('')

                max_word_len = 0
                for word in final_words:
                    if len(word) > max_word_len:
                        max_word_len = len(word)

                # 4 for the leading whitespace and 5 for the '  :  '
                width = 4 + max_word_len + 5
                max_width = 120

                for word in final_words:
                    if word in self.helptext:
                        for (index, line) in enumerate(format_help_text(self.helptext[word], max_width - width)):
                            if index == 0:
                                output.append("    %s  :  %s" % (word.ljust(max_word_len), line))
                            else:
                                output.append("    %s     %s" % (''.ljust(max_word_len), line))
                    else:
                        re_number_range = re.match('<(\d+)-(\d+)>', word)

                        if re_number_range:
                            output.append("    %s  :  An integer from %s to %s" % (word.ljust(max_word_len), re_number_range.group(1), re_number_range.group(2)))
                        else:
                            output.append("    %s  :  add help text" % word.ljust(max_word_len))

                if found_optional_branch:
                    output.append('    <ENTER>')

                if self.daemon:
                    # If we are running as a daemon it will be up to the client to run
                    # get_bash_prompt() so that we get the client's bash prompt instead
                    # of the deamon's bash prompt.
                    print '\n'.join(output)
                else:
                    # If we are here then we are printing TAB complete options,
                    # with help text, to STDERR
                    cmd = self.argv[0]
                    argv = self.argv[1:]

                    if argv:
                        output.append('%s%s %s%s' % (get_bash_prompt(), cmd, ' '.join(argv), ' ' if ended_with_space else ''))
                    else:
                        output.append("%s%s " % (get_bash_prompt(), cmd))

                    sys.stderr.write('\n'.join(output))

        else:
            raise Exception("%s is an invalid options_style" % options_style)
